##################################################################
#
# (C) Copyright 2006 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of Bling.
#
# Bling is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Bling is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Bling; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################
"""
PJS -- (pee jays)

A Python implementation of Ruby on Rails' RJS system.

This allows the expression of Prototype.js javascript code
from native Python commands.

@author: Benjamin Saller <bcsaller@objectrealms.net>
@copyright: ObjectRealms, LLC. 2006
@license: GNU Public License, Version 2.1+

"""
# This is _very_ much like the RJS generation system,
# but using Python infrastructure. Interestingly they have more
# callbacks available to them and can provide more syntactic sugar
# than I can here. The 'with' statement of Python 2.5 opens some
# doors for things like with delay(10): xxx
# but they have a nice assignment syntax I can't match
import json

def nothing(*args, **kwargs): return None

def options_for_javascript(dict):
    return "{%s}" % (', '.join("%s:%s" % (json.write(k),
                                          json.write(v)) for k,v in dict.iteritems()))

def arg_for_js(arg):
    try:
        return json.write(arg)
    except json.WriteException:
        # we just tried to serialize a jsproxy object
        # we want to patch up a native js response
        # that should eval ok on the client
        return arg.tojs()
    
def arguments_for_call(*args): return ', '.join([arg_for_js(a) for a in args])
def add_options_for_javascript(jsdict):
    if not jsdict: return ''
    return ", %s" % (options_for_javascript(jsdict))


class JavascriptPage(object):
    # _data  is the output buffer
    # _ns tracks variables in the namespace
    def __init__(self, buf=None):
        self._data = []
        self._ns = {}
        if buf:self.write(buf)

    # Data
    @property
    def lines(self): return self._data
    @property
    def last(self):
        if self.lines: return self.lines[-1]
        return ''
    
    @property
    def page(self): return self


    # Raw I/O
    def write(self, msg="\n"): self.lines.append(msg)
    def __iadd__(self, other): self.write(other)
    def __str__(self): return '\n'.join(self.lines)
    def __nonzero__(self): return bool(self._data)

    def getvalue(self):
        # collapse the lines
        self._data = [str(self)]
        return str(self)

    # Mid-level I/O
    def append(self, call):
        if not self.lines:
            self.record(call)
        else:
            self.lines[-1] = "%s.%s;" %(
                self.last.rstrip(";"), call)

    def __irshift__(self, msg): self.append(msg)

    def record(self, line):
        line = "%s;" % str(line).rstrip(";")
        self.write(line)

    def update_last(self, line):
        if self.lines: self.lines[-1] = line
        else: self.lines.append(line)

    # Script output
    def tag(self):
        """Enclose the output in a script tag"""
        return '''<script type="text/javascript">//<![CDATA[\n%s\n//]]></script>''' % str(self)

    # Proxy interface
    # page[el] -> jselement
    # page.foo -> ref to js variable 
    def __getattr__(self, key): return jsvariable(self, key)
    def __getitem__(self, key): return jselement(self, key)
    def select(self, pattern): return jscollectionelement(self, pattern)

    # Workhorse
    def __call__(self, function, *args, **kwargs):
        args = arguments_for_call(*args)
        options = add_options_for_javascript(kwargs)

        if not args and options: options = options[1:]
        meth = "%s(%s%s)" % (function, args, options)
        self.record(meth)
        return nothing

    def assign(self, variable, value):
        arg = arg_for_js(value)
        self.record("%s = %s" %(variable, arg))


    # Utility
    def redirect_to(self, location): self.assign('window.location.href', location)

    def insert_html(self, position, id, **kwargs):
        insertion = position.title()
        self('new Insertion.%s' % insertion,
             id, **kwargs)

    def hide(self, *ids): self('Element.hide', ids)
    def show(self, *ids): self('Element.show', ids)
    def alert(self, message): self('alert', message)

    def remove(self, *args):
        self.record(arg_for_js(args) + ".each('Element.remove')")

    def replace(self, id, value, **kwargs):
        self('Element.replace', id, value, **kwargs)

    def new(self, klass, *args, **kwargs):
        el = kwargs.pop('el', self._getElementForClass(klass))
        args = arguments_for_call(*args)
        options = add_options_for_javascript(kwargs)

        if not args and options: options = options[1:]
        self.record('var %s = new %s(%s%s)' %(el, klass, args, options))
            
        return jsvariable(self, el)

    def _getElementForClass(self, klass):
        d = self._ns.setdefault(klass, 0)
        d += 1
        self._ns[klass] = d
        return "%s%d" % (klass.lower(), d)
    

class jsproxy(object):
    # generator is something with a write method
    # and a lines interface
    def __init__(self, generator, root=None, append=False):
        self.__dict__['generator'] = generator
        if root:
            if append: generator.append(root)
            else: generator.record(root)

    # Workhorse
    def __call__(self, function=None, *args, **kwargs):
        args = arguments_for_call(*args)
        options = add_options_for_javascript(kwargs)

        if not args and options: options = options[1:]
        if function: self.append("%s(%s%s)" % (function, args, options))
        else: self.append("(%s%s)" % (args, options))
        return self

    # Proxy Interface
    def __getattr__(self, key): return jselement(self.generator, key)

    # I/O
    def append(self, call):
        if not call.startswith("("): call = ".%s" % call
        self.generator.lines[-1] = "%s%s;" %(
            self.generator.last.rstrip(";"),
            call)
        
    def record(self, line): self.generator.record(line)


class jselement(jsproxy):
    """A Javascript element ala $('foo')"""

    def __init__(self, generator, id):
        self.id = id
        super(jselement, self).__init__(generator, self.tojs())
        
    def replace_html(self, *args, **kwargs): return self('update', *args, **kwargs)
    update = replace_html

    def replace(self, *args, **kwargs): return self('replace', *args, **kwargs)
    def appear(self, *args, **kwargs): return self('appear', *args, **kwargs)
    def show(self, *args, **kwargs): return self('show', *args, **kwargs)
    def visualEffect(self, *args, **kwargs): return self('visualEffect', *args, **kwargs)

    # Proxy Interface
    def __getattr__(self, key):
        super(jselement, self).__call__(key)
        return nothing

    def tojs(self): return "$(%s)" % arg_for_js(self.id)


class jsvariable(jsproxy):
    """represent a native js variable living on the client.
    when something returns a variable as in

    >>> page = JavascriptPage()
    >>> x = page.foo
    >>> a = x.select('a')
    >>> page.last == 'var _v_1 = foo.select("a");'
    True

    When used in expression they work as follows:

    >>> s = x.selected()
    >>> page.last == 'var _v_2 = foo.selected();'
    True
    >>> node = page.tree.walkPath(s)
    >>> page.last
    'var _v_3 = tree.walkPath(_v_2);'

    Method can return jsvariables w/o needing to pass in  variable
    name. In that case a new _v_ name is allocated.
    """
    
    def __init__(self, generator, variable=None):
        self.generator = generator
        if not variable: variable = generator._getElementForClass("_v_")
        self.variable = variable 
        super(jsvariable, self).__init__(generator)

    def tojs(self): return self.variable

    def _get(self):
        self.append(self.tojs())

    def __getattr__(self, key):
        # its like a closure
        self.key = key
        if self.generator.last and \
               not self.generator.last.startswith(self.variable) and \
               not self.generator.last.startswith("var"):
            self.generator.update_last("%s = %s" % (self.variable,
                                                    self.generator.last))
        # its odd to return self, but we are using this for "bound"
        # methods
        ## XXX: for morning try to patch around this, we might need
        ## that stack back, but we should be returning new vars here,
        ## not cached ones
        ## this can let us do attrs or, on call add (params)
        
        return self

    def __call__(self, *args, **kwargs):
        args = arguments_for_call(*args)
        options = add_options_for_javascript(kwargs)

        if not args and options: options = options[1:]
        v = jsvariable(self.generator)
        meth = "var %s = %s.%s(%s%s)" % (v.variable, self.variable, self.key, args, options)
        self.record(meth)
        return v

        
class jscollection(jsproxy):
    enumerable_methods_with_return = ['all', 'any', 'collect', 'map',
                                      'detect', 'find', 'find_all', 'select', 'max', 'min', 'partition',
                                      'reject', 'sort_by', "grep"]
    enumerable_methods = enumerable_methods_with_return + ['each']

    def __init__(self, generator, pattern=None):
        self.pattern = pattern
        super(jscollection, self).__init__(generator)

    def __getattr__(self, method):
        if method in self.enumerable_methods:
            return self.jsenum(self.generator,
                             self.pattern,
                             method,
                             self)
        else:
            return jsproxy(self.generator, method, append=True)
            return nothing

    def grep(self, pattern):
        self.append("grep(%s)" %( arg_for_js(pattern)))
        return jscollection(self.generator);
        
    class jsenum(jsproxy):
        # Prepare for an enumeration, these require a special new
        # proxy prepared to act on a collection
        # XXX: this is far less flexible than the ruby version
        # but should be generally sufficient
        # for simple things. I think an expanded API here that makes
        # it easy to pass raw js into the enum would address the more
        # complex cases. Thats a concession that avoiding JS 100% of
        # the time is not the goal, more of an 80/20, 90/10 kinda
        # thing.
        def __init__(self, generator, pattern, method, collection):
            self.generator = generator
            self.pattern  = pattern
            self.method = method
            self.collection = collection

        def __call__(self, *args, **kwargs):
            call = self.pattern
            # This is a mess to support the "grep" case,
            # methods returning arrays
            if args: call = args[0]
            if args[1:]: args = arguments_for_call(args[1:])
            else: args = ''
            options = add_options_for_javascript(kwargs)
            if not args and options: options = options[1:]
            meth = "%s(function(el) {el.%s(%s%s)};)" % (self.method,
                                                        call, args, options)
            self.append(meth)
            if self.method in jscollection.enumerable_methods_with_return:
                return jsvariable(self.generator, 'el')



class jscollectionelement(jscollection):
    def __init__(self, generator, pattern):
        self.pattern = pattern
        # note that I call jscollection here
        super(jscollection, self).__init__(generator,
                                            "$$(%s)" %
                                            arg_for_js(pattern))



if __name__ == "__main__":
    import doctest
    import pdb
    import sys
    import traceback
    # When dealing with recursive proxies
    # its nice to have a debugging safety net
    sys.setrecursionlimit(35)
    options = doctest.ELLIPSIS
    doctest.testfile("using_pjs.txt", optionflags=options)
